import lark
import pathlib

from treadi.data import Issue

GRAMMAR = (pathlib.Path(__file__).parent.resolve() / "filter.lark").read_text()
PARSER = lark.Lark(GRAMMAR, parser="lalr")


class Requirement:

    def __call__(self, issue_or_pr) -> bool:
        """Return True if the requirement is met."""
        raise NotImplementedError

    def invert(self, issue_or_pr) -> bool:
        """Return False if the requirement is met."""
        # Some requirments ignore issues,
        # so the inverse must also return True for issues
        return not self(issue_or_pr)


class RequirePR(Requirement):
    """Return True if the issue is a PR."""

    def __call__(self, issue_or_pr) -> bool:
        return issue_or_pr.is_pr


class RequireIssue(Requirement):
    """Return True if the issue is an issue."""

    def __call__(self, issue_or_pr) -> bool:
        return not issue_or_pr.is_pr


class RequireDraftIfPR(Requirement):
    """Return True if it is an Issue or a draft PR."""

    def __call__(self, issue_or_pr) -> bool:
        if issue_or_pr.is_pr:
            return issue_or_pr.draft
        return True

    def invert(self, issue_or_pr) -> bool:
        if issue_or_pr.is_pr:
            return not issue_or_pr.draft
        return True


class RequireApprovedIfPR(Requirement):
    """Return True if it is an Issue or an approved PR."""

    def __call__(self, issue_or_pr) -> bool:
        if issue_or_pr.is_pr:
            return issue_or_pr.approved
        return True

    def invert(self, issue_or_pr) -> bool:
        if issue_or_pr.is_pr:
            return not issue_or_pr.approved
        return True


class RequireChangesRequestedIfPR(Requirement):
    """Return True if it is an Issue or a PR with changes requested."""

    def __call__(self, issue_or_pr) -> bool:
        if issue_or_pr.is_pr:
            return issue_or_pr.changes_requested
        return True

    def invert(self, issue_or_pr) -> bool:
        if issue_or_pr.is_pr:
            return not issue_or_pr.changes_requested
        return True


class RequireNoReviewIfPR(Requirement):
    """Return True if it is an Issue or a PR with no reviews."""

    def __call__(self, issue_or_pr) -> bool:
        if issue_or_pr.is_pr:
            return not issue_or_pr.changes_requested and not issue_or_pr.approved
        return True

    def invert(self, issue_or_pr) -> bool:
        if issue_or_pr.is_pr:
            return issue_or_pr.changes_requested or issue_or_pr.approved
        return True


class RequireAuthor(Requirement):

    def __init__(self, author):
        self._author = author

    def __call__(self, issue_or_pr):
        """Return True if the issue is authored by this author."""
        return self._author.lower() == issue_or_pr.author.lower()


class RequireRepo(Requirement):

    def __init__(self, owner, name):
        self._owner = owner
        self._name = name

    def __call__(self, issue_or_pr):
        """Return True if the issue belongs to the given repo."""
        return (
            self._owner.lower() == issue_or_pr.repo.owner.lower()
            and self._name.lower() == issue_or_pr.repo.name.lower()
        )


class RequireOrg(Requirement):

    def __init__(self, org):
        self._org = org

    def __call__(self, issue_or_pr):
        """Return True if the issue belongs to the given organization."""
        return self._org.lower() == issue_or_pr.repo.owner.lower()


class RequireBaseBeginsWith(Requirement):

    def __init__(self, base_prefix):
        self._base_prefix = base_prefix

    def __call__(self, issue_or_pr) -> bool:
        if issue_or_pr.is_pr:
            return issue_or_pr.base_ref.startswith(self._base_prefix)
        return True

    def invert(self, issue_or_pr) -> bool:
        if issue_or_pr.is_pr:
            return not self(issue_or_pr)
        return True


class Invert(Requirement):

    def __init__(self, requirement):
        self._requirement = requirement

    def __call__(self, issue_or_pr):
        """Return the opposite of what the requirement returns."""
        return self._requirement.invert(issue_or_pr)


class RequireAll(Requirement):

    def __init__(self, requirements):
        self._requirements = tuple(requirements)

    def __call__(self, issue_or_pr):
        """Return True iff all requirements are met."""
        for requirement in self._requirements:
            if not requirement(issue_or_pr):
                return False
        return True


class FilterTransformer(lark.Transformer):
    """Turn a lark parse tree into a callable that filters issues."""

    def start(self, args):
        return RequireAll(args)

    def statement(self, args):
        return args[0]

    def maybe_negated_statement(self, args):
        if args[0] == "-":
            return Invert(args[1])
        return args[0]

    def type_stmt(self, args):
        if args[0] == "pr":
            return RequirePR()
        return RequireIssue()

    def author_stmt(self, args):
        return RequireAuthor(args[0].value)

    def assignee_stmt(self, args):
        raise NotImplementedError

    def is_stmt(self, args):
        if args[0] == "pr":
            return RequirePR()
        if args[0] == "issue":
            return RequireIssue()
        if args[0] == "draft":
            return RequireDraftIfPR()

    def review_stmt(self, args):
        if args[0] == "approved":
            return RequireApprovedIfPR()
        if args[0] == "changes_requested":
            return RequireChangesRequestedIfPR()
        if args[0] == "none":
            return RequireNoReviewIfPR()

    def repo_stmt(self, args):
        return RequireRepo(args[0].value, args[1].value)

    def org_stmt(self, args):
        return RequireOrg(args[0].value)

    def base_stmt(self, args):
        return RequireBaseBeginsWith(args[0].value)


def parse(filter: str):
    tree = PARSER.parse(filter)
    return FilterTransformer().transform(tree)



import copy

from treadi.data import Issue
from treadi.data import PullRequest
from treadi.data import Repository
import treadi.filter as tf

REPO_ROS_ROSDITRO = Repository(
    owner="ros",
    name="rosdistro",
)
REPO_ROS2_ROS2 = Repository(
    owner="ros2",
    name="ros2",
)
ISSUE_ROS_ROSDISTRO = Issue(
    repo=REPO_ROS_ROSDITRO,
    author="sloretz",
    created_at=None,
    updated_at=None,
    number=42,
    title="fix stuff",
    url="",
    is_pr=False,
)
PR_ROS_ROSDISTRO = PullRequest(
    repo=REPO_ROS_ROSDITRO,
    author="ghost",
    created_at=None,
    updated_at=None,
    number=1,
    title="break stuff",
    url="",
    is_pr=True,
    base_ref="rolling",
)
DRAFT_PR_ROS_ROSDISTRO = PullRequest(
    repo=REPO_ROS_ROSDITRO,
    author="ghost",
    created_at=None,
    updated_at=None,
    number=1,
    title="break stuff",
    url="",
    is_pr=True,
    draft=True,
    base_ref="main",
)
APPROVED_PR_ROS_ROSDISTRO = PullRequest(
    repo=REPO_ROS_ROSDITRO,
    author="ghost",
    created_at=None,
    updated_at=None,
    number=1,
    title="break stuff",
    url="",
    is_pr=True,
    approved=True,
    base_ref="master",
)
CHANGES_REQUESTED_PR_ROS_ROSDISTRO = PullRequest(
    repo=REPO_ROS_ROSDITRO,
    author="ghost",
    created_at=None,
    updated_at=None,
    number=1,
    title="break stuff",
    url="",
    is_pr=True,
    changes_requested=True,
    base_ref="rolling",
)


def test_require_pr():
    for f in [tf.parse("is:pr"), tf.parse("type:pr")]:
        assert f(PR_ROS_ROSDISTRO)
        assert not f(ISSUE_ROS_ROSDISTRO)


def test_require_issue():
    for f in [tf.parse("is:issue"), tf.parse("type:issue")]:
        assert f(ISSUE_ROS_ROSDISTRO)
        assert not f(PR_ROS_ROSDISTRO)


def test_require_author():
    f = tf.parse("author:sloretz")
    assert f(ISSUE_ROS_ROSDISTRO)
    f = tf.parse("author:bob")
    assert not f(ISSUE_ROS_ROSDISTRO)


def test_require_repo():
    assert tf.RequireRepo("ros", "rosdistro")(ISSUE_ROS_ROSDISTRO)
    assert not tf.RequireRepo("ros2", "ros2")(ISSUE_ROS_ROSDISTRO)


def test_require_draft_if_pr():
    f = tf.parse("is:draft")
    assert f(ISSUE_ROS_ROSDISTRO)
    assert not f(PR_ROS_ROSDISTRO)
    assert f(DRAFT_PR_ROS_ROSDISTRO)


def test_require_base_if_pr():
    f = tf.parse("base:roll")
    fnot = tf.parse("-base:roll")

    rolling_pr = copy.deepcopy(PR_ROS_ROSDISTRO)
    main_pr = copy.deepcopy(PR_ROS_ROSDISTRO)
    rolling_pr.base_ref = "rolling"
    main_pr.base_ref = "main"

    assert f(ISSUE_ROS_ROSDISTRO)
    assert fnot(ISSUE_ROS_ROSDISTRO)
    assert f(rolling_pr)
    assert not fnot(rolling_pr)
    assert not f(main_pr)
    assert fnot(main_pr)


def test_invert_require_draft_if_pr():
    f = tf.parse("-is:draft")
    assert f(ISSUE_ROS_ROSDISTRO)
    assert f(PR_ROS_ROSDISTRO)
    assert not f(DRAFT_PR_ROS_ROSDISTRO)


def test_require_approved_if_pr():
    f = tf.parse("review:approved")
    assert f(ISSUE_ROS_ROSDISTRO)
    assert not f(PR_ROS_ROSDISTRO)
    assert f(APPROVED_PR_ROS_ROSDISTRO)


def test_invert_require_approved_if_pr():
    f = tf.parse("-review:approved")
    assert f(ISSUE_ROS_ROSDISTRO)
    assert f(PR_ROS_ROSDISTRO)
    assert not f(APPROVED_PR_ROS_ROSDISTRO)


def test_require_changes_requested_if_pr():
    f = tf.parse("review:changes_requested")
    assert f(ISSUE_ROS_ROSDISTRO)
    assert not f(PR_ROS_ROSDISTRO)
    assert f(CHANGES_REQUESTED_PR_ROS_ROSDISTRO)


def test_invert_require_changes_requested_if_pr():
    f = tf.parse("-review:changes_requested")
    assert f(ISSUE_ROS_ROSDISTRO)
    assert f(PR_ROS_ROSDISTRO)
    assert not f(CHANGES_REQUESTED_PR_ROS_ROSDISTRO)


def test_require_no_review_if_pr():
    f = tf.parse("review:none")
    assert f(ISSUE_ROS_ROSDISTRO)
    assert f(PR_ROS_ROSDISTRO)
    assert not f(APPROVED_PR_ROS_ROSDISTRO)
    assert not f(CHANGES_REQUESTED_PR_ROS_ROSDISTRO)


def test_multiple():
    f = tf.parse("author:sloretz is:issue")
    assert f(ISSUE_ROS_ROSDISTRO)
    assert not f(PR_ROS_ROSDISTRO)



import logging
import webbrowser

from kivy.animation import Animation
from kivy.app import App
from kivy.core.window import Window
from kivy.properties import ColorProperty
from kivy.properties import ObjectProperty
from kivy.uix.behaviors import ButtonBehavior
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from kivy.uix.screenmanager import Screen

from ..data import Issue
from ..data import is_same_issue
from ..filter import parse as parse_filter


class IssueScreenGoBackPopup(Popup):

    def on_go_back(self):
        App.get_running_app().switch_to_pick_repos(direction="right")
        self.dismiss(animation=False)


class ChangesRequestedWidget(Label):
    pass


class ApprovedWidget(Label):
    pass


class DraftWidget(Label):
    pass


class IssueWidget(ButtonBehavior, BoxLayout):

    color = ColorProperty(defaultvalue=[0.6, 0.6, 0.6, 1])

    issue = ObjectProperty(
        Issue(),
        rebind=True,
    )

    def __init__(self, issue, dismiss_callback, **kwargs):
        self.issue = issue
        self.dismiss_callback = dismiss_callback
        super().__init__(**kwargs)

        if issue.is_pr:
            if issue.draft:
                self.ids.pr_status.add_widget(DraftWidget())
            elif issue.approved:
                self.ids.pr_status.add_widget(ApprovedWidget())
            elif issue.changes_requested:
                self.ids.pr_status.add_widget(ChangesRequestedWidget())

    def on_press(self):
        if self.last_touch.button == "left":
            self.color = [0.6, 0.6, 0.8, 1]

    def on_release(self):
        # Oh yeah, the whole point is to open the issue
        if self.last_touch.button == "left":
            webbrowser.open(self.issue.url)

    def do_dismiss_callback(self):
        # Only dismiss once
        if self.dismiss_callback is not None:
            d = self.dismiss_callback
            self.dismiss_callback = None
            d(self)


class IssueScreen(Screen):

    def __init__(self, *args, **kwargs):
        self._popup = None
        self._filter = None
        self._logger = logging.getLogger("IssueScreen")
        if "name" not in kwargs:
            kwargs["name"] = "issues-screen"
        super().__init__(*args, **kwargs)

    def validate_filter(self):
        # Called when on_validate_text is called on filter TextInput
        filter_str = self.ids.filter.text
        if filter_str.strip() == "":
            self._filter = None
            self.ids.filter.background_color = (1, 1, 1, 1)
        else:
            try:
                self._filter = parse_filter(filter_str.strip())
                self.ids.filter.background_color = (0.9, 1, 0.9, 1)
            except:
                self._logger.exception("Cannot parse filter")
                self._filter = None
                self.ids.filter.background_color = (1, 0.5, 0.5, 1)

        self._refresh_issues()

    def on_text_changing(self):
        # called hen on_text is called on filter TextInput
        # Filters don't apply until the user hits enter, so display as yellow
        self.ids.filter.background_color = (1, 1, 0.9, 1)

    def on_pre_enter(self):
        issue_cache = App.get_running_app().issue_cache
        self._refresh_issues()
        Window.bind(on_key_down=self.on_key_down)

    def on_pre_leave(self):
        Window.unbind(on_key_down=self.on_key_down)

    def _forget_popup(self, _):
        self._popup = None

    def on_key_down(self, window, key, scancode, codepoint, modifiers):
        if key == 27:  # ESCAPE KEY
            # Loading issues again is time consuming.
            # Make the user confirm that they really do want to pick
            # repos again.
            if self._popup is None:
                self._popup = IssueScreenGoBackPopup()
                self._popup.bind(on_dismiss=self._forget_popup)
                self._popup.open(animation=False)
            else:
                self._popup.dismiss(animation=False)

    def _refresh_issues(self):
        # Clear and re-add issues
        self.ids.stack.clear_widgets()
        issue_cache = App.get_running_app().issue_cache
        issues = issue_cache.most_recent_issues(n=5, filter=self._filter)
        for i in issues:
            self._add_issue(i)

    def _add_issue(self, issue):
        self.ids.stack.add_widget(IssueWidget(issue, self.dismiss))

    def dismiss(self, issue_widget):
        issue = issue_widget.issue
        issue_cache = App.get_running_app().issue_cache

        issue_cache.dismiss(issue)

        child_issues = []
        for child in self.ids.stack.children:
            if hasattr(child, "issue"):
                child_issues.append(child.issue)

        # aim for 1 additional issue, because one is being dismissed
        consider_issues = issue_cache.most_recent_issues(
            n=1 + len(child_issues), filter=self._filter
        )
        for itc in consider_issues:
            displaying_issue = False
            for chi in child_issues:
                if is_same_issue(itc, chi):
                    displaying_issue = True
                    break
            if not displaying_issue:
                self._add_issue(itc)
                break

        # Animate the dismissed widget shrinking, so new issue reveals from below
        anim = Animation(
            size_hint_y=0, opacity=0, duration=0.125, transition="out_cubic"
        )
        anim.bind(on_complete=lambda *args: self.ids.stack.remove_widget(issue_widget))
        anim.start(issue_widget)



from kivy.app import App
from kivy.clock import Clock
from kivy.properties import NumericProperty
from kivy.uix.screenmanager import Screen

from ..issue_loader import IssueLoader
from .issue_screen import IssueScreen


class IssueLoadingScreen(Screen):

    progress = NumericProperty(0.0)

    def __init__(self, repos, **kwargs):
        App.get_running_app().issue_loader = IssueLoader(
            App.get_running_app().gql_client,
            repos,
            App.get_running_app().issue_cache,
            self.update_progress,
        )
        if "name" not in kwargs:
            kwargs["name"] = "issue-loading-screen"
        super().__init__(**kwargs)

    def update_progress(self, progress):
        self.progress = progress * 100
        if progress >= 1.0:
            Clock.schedule_once(lambda dt: self.switch_to_issues())

    def switch_to_issues(self):
        # Must only be called on main thread
        self.manager.switch_to(IssueScreen(), direction="left")



from importlib.metadata import version

from kivy.app import App
from kivy.core.window import Window
from kivy.uix.button import Button
from kivy.uix.popup import Popup
from kivy.uix.screenmanager import Screen

from .. import config
from ..repo_loader import parse_repository_list
from .repo_loading_screen import RepoLoadingScreen


class RepoButton(Button):
    """RepoPickerScreen button.

    This is a class so I can customize in kv language.
    """

    def __init__(self, repo_picker, *args, **kwargs):
        self._repo_picker = repo_picker
        self._display_name = kwargs["text"]
        super().__init__(*args, **kwargs)

    def on_release(self):
        self._repo_picker.use_repo_list(self._display_name)


class RepoPickerScreenHelpPopup(Popup):

    def __init__(self, config):
        self._version = version("TreadI")
        self._config = config
        super().__init__()


class RepoPickerScreen(Screen):

    def __init__(self, *args, **kwargs):
        self._popup = None
        self._config = config.Config()
        if "name" not in kwargs:
            kwargs["name"] = "repo-picker-screen"
        super().__init__(*args, **kwargs)

    def on_pre_enter(self):
        for name in self._config.repository_list_names():
            btn = RepoButton(repo_picker=self, text=name)
            self.ids.stack.add_widget(btn)
        Window.bind(on_key_down=self.on_key_down)

    def on_pre_leave(self):
        Window.unbind(on_key_down=self.on_key_down)

    def on_key_down(self, window, key, scancode, codepoint, modifiers):
        if key == 47 and "shift" in modifiers:  # ? KEY
            if self._popup is None:
                self._popup = RepoPickerScreenHelpPopup(self._config)
                self._popup.bind(on_dismiss=self._forget_popup)
                self._popup.open(animation=False)
                print("Opening popup")
            else:
                self._popup.dismiss(animation=False)

    def _forget_popup(self, _):
        self._popup = None

    def use_repo_list(self, display_name):
        repo_list = self._config.repository_list(display_name)
        repo_loader = parse_repository_list(
            repo_list.content,
            # TODO pass this into repo loader
            gql_client=App.get_running_app().gql_client,
        )
        self.manager.switch_to(RepoLoadingScreen(repo_loader), direction="left")
        App.get_running_app().title += f": {display_name}"



from .data import is_same_issue
from threading import Lock


class IssueCache:

    def __init__(self):
        self.__upcomming = []
        self.__dismissed = []
        self.__lock = Lock()

    def insert(self, issue):
        """Insert an issue into the cache.

        If the cache already has newer info for the issue,
        it will silently ignore this insertion.
        """
        with self.__lock:
            self._insert(issue)

    def _insert(self, issue):
        for di, d in enumerate(self.__dismissed):
            if is_same_issue(issue, d):
                if issue.updated_at <= d.updated_at:
                    # Not new data, nothing to do here
                    return
                # Put it into the incomming list
                del self.__dismissed[di]
                break
        # If we get here, the issue belongs in the
        # upcomming list
        for ui, u in enumerate(self.__upcomming):
            if is_same_issue(issue, u):
                if issue.updated_at > u.updated_at:
                    self.__upcomming[ui] = issue
                    return
                return
        # If we get here, the issue is new to us
        self.__upcomming.append(issue)

    def dismiss(self, issue):
        """
        Dismiss an issue so that it no longer
        comes up in `most_recent_not_dismissed`.
        """
        with self.__lock:
            self._dismiss(issue)

    def _dismiss(self, issue):
        for di, d in enumerate(self.__dismissed):
            if is_same_issue(issue, d):
                # already dismissed, nothing to do!
                return
        for ui, u in enumerate(self.__upcomming):
            if is_same_issue(issue, u):
                # Move from upcomming to dismiseed
                del self.__upcomming[ui]
                self.__dismissed.append(u)
                return

    def most_recent_issues(self, n=1, filter=None):
        """
        Return the n most recently updated and not
        dismissed issues.
        """
        with self.__lock:
            return self._most_recent_issues(n, filter)

    def newest_update_time(self):
        with self.__lock:
            self._sort()
            u = self.__upcomming[:1]
            d = self.__dismissed[:1]
        if not u and not d:
            return None
        if not u:
            return d[0].updated_at
        if not d:
            return u[0].updated_at
        if d[0].updated_at > u[0].updated_at:
            return d[0].updated_at
        return u[0].updated_at

    def _most_recent_issues(self, n, filter):
        self._sort()
        if filter is None:
            return self.__upcomming[:n]
        issues = []
        for i in self.__upcomming:
            if filter(i):
                # True means it passes the filter
                issues.append(i)
            if len(issues) >= n:
                break
        return issues

    def _sort(self):
        self.__upcomming.sort(reverse=True, key=lambda i: i.updated_at)
        self.__dismissed.sort(reverse=True, key=lambda i: i.updated_at)



import threading

from treadi.data import Repository
from treadi.repo_loader import parse_repository_list


def test_parse_all_directives():
    content = """
    owned-by:@me
    org:ros2
    vcstool:https://foobar.com/some.repos
    my-org/MyRepo
    """
    repo_loader = parse_repository_list(content, gql_client="fake")
    assert len(repo_loader.repo_loaders) == 4


def test_single_repo_directives():
    content = """
    sloretz/baz.foo
    my-org/MyRepo
    cool/story_bro
    """
    repo_loader = parse_repository_list(content, gql_client="fake")
    e = threading.Event()
    repo_loader.begin_loading(lambda _: e.set())
    e.wait()

    assert Repository(owner="sloretz", name="baz.foo") in repo_loader.repos()
    assert Repository(owner="my-org", name="MyRepo") in repo_loader.repos()
    assert Repository(owner="cool", name="story_bro") in repo_loader.repos()
